{% if name %}
    <h4>{{ name }}</h3>
{% endif %}
<table class="table medium-size-table">
    <tbody>
        {% for row in rows %}
        <tr>
            <th class="table-rowheader">{{ row['name'] }}</th>
            <td>{{ row['value'] }}</td>
        </tr>
        {% endfor %}
    </tbody>
</table>
{% if caption %}
    <div class="caption text-center">
        {{ caption }}
    </div>
{% endif %}