<h2>Place analysis</h2>

<div>

    <h3>Visits per tile</h3>
    <p class="info">{{privacy_info}}</p>
    
    <iframe src="{{output_filename}}/maps/visits_per_tile_map.html" width="80%" height="600" style="border:1px solid black;">
    </iframe>
<!---    <div style="width: 800px; height: 600px;">
        {{counts_per_tile_map}}
    </div>
--->
    
    <div>
        {{visits_per_tile_legend}}
    </div>
    <p>
        {{points_outside_tessellation_info}}
    </p>
</div>


<div>
    {{counts_per_tile_summary_table}}
    
    {{counts_per_tile_cumsum_linechart}}
    <p class="info">
        The cumulated sum of visits visualizes the distribution of visits over tiles:
        If all tiles are visited equally, the cumulated sum follows a straight diagonal line (gray line). 
        The larger the share of single tiles in the total number of visits, the steeper the curve.
    </p>
</div>

<div>
    <h4>Ranking of most frequently visited tiles</h4>
    {{most_freq_tiles_ranking}}
</div>
    
<div>
    <h3>Visits per tile and time window</h3>
    
    <p class="info">Each map shows the <strong>arrivals (destinations)</strong> for the respective time window for each tile, split by weekday and weekend, as absolute counts and as the deviation from tile average.<br>    
    
    <p class="info">{{privacy_info}}</p>

        <div class = "map_png">
        {{counts_per_tile_time_map}}
    </div>
</div>