<!doctype html>

<html lang="en">
<link rel="stylesheet" href="{{output_filename}}/assets/bootstrap.min.css">
<link rel="stylesheet" href="{{output_filename}}/assets/style.css">

<head>
<meta charset="utf-8">
<meta name="viewport" content="width=device-width, initial-scale=1">

<title>DP Mobility Report</title>
<meta name="description" content="DP Mobility Report rendered as HTML.">
</head>

<body>
    
<div class="container">
    <h1>DP Mobility Report</h1>

    <div class="segment grey">
        {{config_segment}}
    </div>

    <div class="segment">
        {{overview_segment}}
    </div>

    <div class="segment">
        {{place_analysis_segment}}
    </div>

    <div class="segment">
        {{od_analysis_segment}}
    </div>

    <div class="segment">
        {{user_analysis_segment}}
    </div>
</div>

</body>


</html>