<h2>Overview</h2>

<div>
    {{dataset_stats_table}}
</div>

<div>
    {{missing_values_table}}
</div>

<div>
    <h3>Temporal properties</h3>
    
    <div>
        <h4>Trips over time</h4>
        {{trips_over_time_linechart}}

        <p class ="info">
            {{trips_over_time_info}}
        </p>
    </div>

    {{trips_over_time_summary_table}}
    <p class="info">
        {{trips_over_time_moe_info}}
    </p> 
    
    <div>
        <h4>Trips per weekday</h4>
        {{trips_per_weekday_barchart}}
    </div>
    
    <div>
        <h4>Trips per hour of day split by weekday and weekend</h4>
        {{trips_per_hour_linechart}}
    </div>
</div>
    