{% if name %}
    <h4>{{ name }}</h3>
{% endif %}
<table class="table align-right">
    <tbody>
        <tr>
            {% if rows[0]['rank'] %}
                <th class ="table-header"></th>
            {% endif %}
            <th class ="table-header"></th>
            <th class ="table-header">lower limit</th>
            <th class ="table-header">estimate</th>
            <th class ="table-header">upper limit</th>
        </tr>
        {% for row in rows %}
        <tr>
            {% if row['rank'] %}
                <th class="table-rowheader">{{ row['rank'] }}</th>
            {% endif %}
            <th class = "table-rowheader">{{ row['name'] }}</th>
            <td>{{ row['lower_limit'] }}</td>
            <td class="estimate">{{ row['estimate'] }}</td>
            <td>{{ row['upper_limit'] }}</td>
        </tr>
        {% endfor %}
    </tbody>
</table>
{% if caption %}
    <div class="caption text-center text-muted">
        {{ caption }}
    </div>
{% endif %}